/*************************************************************************\
    Copyright (c) 2010 Corporation for National Research Initiatives;
                        All rights reserved.
     The CNRI open source license for this software is available at 
                  http://hdl.handle.net/1895.26/1014
---------------------------------------------------------------------------
Developer: Sean Reilly <sreilly@cnri.reston.va.us>
\*************************************************************************/

package net.cnri.apps.doserver.operators;


import net.cnri.apps.doserver.*;
import net.cnri.apps.doserver.txnlog.*;
import net.cnri.do_api.*;
import net.cnri.dobj.*;
import net.cnri.knowbots.lib.*;
import net.cnri.knowbots.station.CollaborationHub;
import net.cnri.util.*;

import org.apache.lucene.document.*;
import org.apache.lucene.index.*;
import org.apache.lucene.queryParser.ParseException;
import org.apache.lucene.search.*;
import org.apache.lucene.store.*;
import java.io.*;
import java.util.*;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

/**
 * Operator that provides text indexing capabilities using the Lucene search engine.
 * This version analyzes the transaction log from a particular repository and 
 * indexes any objects in that repository.
 */
public class LuceneIndexer
extends Knowbot
implements DOOperation,
Runnable,
DOConstants
{
    public static final String INDEX_BUILDER_CLASS_KEY = "index_builder_class";

    public static final int MAX_DOCS = 4000; // maximum number of hits returned by a search 
    public static final String DEBUG_PROPERTY = "do.debug_lucene";
    private static final String INDEX_DIR = "index";
    public static final String LUCENE_CONFIG_ELEMENT = "lucene.config";
    public static final String FEDERATED_INDEXES_KEY = "federated_indexes";

    // transaction log constants
    public static final String PARAM_TXN_ID = "txn_id";
    public static final String PARAM_OBJ_ID = "object_id";
    public static final String PARAM_TXN_TYPE = "txn_type";
    public static final String PARAM_TIMESTAMP = "tstamp";
    public static final String PARAM_ACTUAL_TIMESTAMP = "atstamp";
    public static final String PARAM_ELEMENT_ID = "elementid";

    public static final String TXN_TYPE_ADD_OBJ = "add";
    public static final String TXN_TYPE_DEL_OBJ = "del";
    public static final String TXN_TYPE_UPDATE_ELEMENT = "update_element";
    public static final String TXN_TYPE_DEL_ELEMENT = "del_element";
    public static final String TXN_TYPE_COMMENT = "comment";

    private int NUM_OBJECTS_BATCH_SIZE = 1000; // only index this many objects at a time.

    public boolean DEBUG = false;

    private transient volatile boolean keepRunning = false; 

    private final Hashtable myOperations = new Hashtable();
    private final String operationKeys[] = {
            SEARCH_OP_ID,
            REINDEX_OBJECT_ID
    };
    
    private final Main serverMain;
    private final DOAuthentication auth;
    private final File scanStatusFile;
    private final File suppressIndexingFile;
    private final FSDirectory indexDirectory;
    private final IndexBuilder indexBuilder;
    private StorageProxy repoStorage = null; // lazily initialized then immutable
    
    private List federatedSearchTargets = new ArrayList();
    private StreamTable indexScanStatus = null;

    private volatile IndexWriter indexWriter = null;
    private final Object indexWriterLock = new Object();
    
    private Thread shutdownHook;
    private final Thread indexerThread;
    private final TransactionQueueListener queueListener;
    private final ScheduledExecutorService execServ = Executors.newSingleThreadScheduledExecutor();
    private final Object updateWaitCondition = new Object();
    private volatile boolean txnUpdateNeeded = false;
    private volatile boolean periodicUpdateNeeded = true;
    
    private final IndexSearcherManager indexSearcherManager;

    public LuceneIndexer(final Main serverMain) 
    throws Exception
    {
        this.serverMain = serverMain;

        DEBUG = DEBUG || System.getProperty(DEBUG_PROPERTY, "").toLowerCase().startsWith("y");
        for(int i=0; i<operationKeys.length; i++)
            myOperations.put(operationKeys[i], "");
        auth = AbstractAuthentication.getAnonymousAuth();


        // add a shutdown hook
        shutdownHook = new Thread() {
            public void run() {
                try {
                    keepRunning = false;
                    if(queueListener!=null) serverMain.getTxnQueue().removeQueueListener(queueListener);
                    if(execServ!=null) execServ.shutdownNow();
                    IndexWriter im = indexWriter;
                    if(im!=null) im.close();
                    saveSourceStatus();
                } catch (Exception e) {
                    System.err.println("Error closing indexwriter: "+e);
                }
            }
        };
        Runtime.getRuntime().addShutdownHook(shutdownHook);

        String repoID = serverMain.getServerID();
        Storage storage = serverMain.getStorage();
        File serverBase = serverMain.getBaseFolder();
        File indexBase = new File(serverBase, INDEX_DIR);
        suppressIndexingFile = new File(indexBase, "suppress_indexing");
        File indexDir = new File(indexBase, "index");
        
        String indexBuilderClassName = serverMain.getConfigVal(INDEX_BUILDER_CLASS_KEY);
        if(indexBuilderClassName!=null) {
            indexBuilder = (IndexBuilder) Class.forName(indexBuilderClassName).getConstructor().newInstance();
        }
        else {
            indexBuilder = new DefaultIndexBuilder();
        }
        indexBuilder.init(serverMain);

        indexDirectory = FSDirectory.open(indexDir);
        if(!IndexReader.indexExists(indexDirectory)) {
            IndexWriter writer = new IndexWriter(indexDirectory, indexBuilder.getAnalyzer(), true, IndexWriter.MaxFieldLength.UNLIMITED);
            writer.close();
        }
        indexSearcherManager = new IndexSearcherManager(IndexReader.open(indexDirectory,true));

        try {
            indexScanStatus = new StreamTable();
            scanStatusFile = new File(indexBase, "status.dct");
            if(scanStatusFile.exists()) {
                indexScanStatus.readFromFile(scanStatusFile);
            }
        } catch (StringEncodingException e) {
            throw new IOException("Error reading lucene scan status: "+e);
        }

        // kick off the asynchronous indexer..
        indexerThread = new Thread(this);
        indexerThread.setPriority(Thread.MIN_PRIORITY);
        indexerThread.start();

        execServ.scheduleWithFixedDelay(new Runnable() {
            public void run() {
                periodicUpdateNeeded = true;
                synchronized(updateWaitCondition) {
                    updateWaitCondition.notify();
                }
            }
        },30,30,TimeUnit.SECONDS);

        queueListener = new LuceneTransactionQueueListener();
        serverMain.getTxnQueue().addQueueListener(queueListener);
    }

    private class LuceneTransactionQueueListener implements TransactionQueueListener {
        public void transactionAdded(Transaction txn) {
            if(!txnUpdateNeeded) {
                txnUpdateNeeded = true;
                synchronized(updateWaitCondition) {
                    updateWaitCondition.notify();
                }
            }
        }
    }

    public void executeInContext() {
        try {
            indexerThread.join();
        } catch (Exception e) {
            System.err.println("error waiting for indexer thread in LuceneIndexer.executeInContext(): "+e);
            e.printStackTrace(System.err);
        }
    }


    public boolean canHandleOperation(DOOperationContext context) {
        return myOperations.containsKey(context.getOperationID().toLowerCase());
    }


    /**
     * Returns a list of operations that this operator can perform
     * on the object identified by the DOOperationContext parameter.
     */
    public String[] listOperations(DOOperationContext context) {
        // we always build an index in the main repository object
        if(context.getTargetObjectID().equals(context.getServerID())) {
            return operationKeys;
        }

        try {
            if(context.getStorage().doesDataElementExist(LUCENE_CONFIG_ELEMENT))
                return operationKeys;
        } catch (Exception e) {
            System.err.println("Error checking for element: element="+
                    LUCENE_CONFIG_ELEMENT+" error="+e);
        }

        return null;
    }


    public void run() {
        keepRunning = true;
        System.err.println("Starting Lucene Indexer async indexing");

        try {
            while(keepRunning) {
                boolean bigUpdate = periodicUpdateNeeded;
                try {
                    updateIndex(bigUpdate);
                } catch (Exception t) {
                    System.err.println("Error updating index: "+t);
                    t.printStackTrace(System.err);
                }
                if(bigUpdate) periodicUpdateNeeded = false;

                synchronized(updateWaitCondition) {
                    while(!txnUpdateNeeded && !periodicUpdateNeeded) {
                        try {
                            updateWaitCondition.wait();
                        }
                        catch(InterruptedException e) {
                            break;
                        }
                    }
                }
            }
        } finally {
            Thread shutdown = this.shutdownHook;
            if(shutdown!=null) {
                Runtime.getRuntime().removeShutdownHook(shutdownHook);
                shutdown.run();
                this.shutdownHook = null;
            }
        }
    }


    /**
     * Performs the given operation (which this object has advertised that it
     * can handle) which consists of reading input (if any is expected) from the
     * given InputStream and writing the output of the operation (if any) to the
     * OutputStream.  This method should *always* close the input and output streams
     * when finished with them.  If there are any errors in the input, the error
     * message must be communicated on the OutputStream since all errors must be
     * at the application level.  Any exceptions thrown by this method will *not*
     * be communicated to the caller and are therefore not acceptable.
     */
    public void performOperation(DOOperationContext context, InputStream in, OutputStream out) {
        try {
            String operation = context.getOperationID();
            if(operation.equalsIgnoreCase(SEARCH_OP_ID)) {
                doSearch(context, in, out);
            } else if(operation.equalsIgnoreCase(REINDEX_OBJECT_ID)) {
              doReindex(context, in, out);
            } else {
                // shouldn't happen
                sendErrorResponse(out, "Operation '"+operation+"' not implemented!",
                        DOException.OPERATION_NOT_AVAILABLE);
                return;
            }
        } catch (Exception e) {
            e.printStackTrace(System.err);
        } finally {
            try { in.close(); } catch (Exception e) {}
            try { out.close(); } catch (Exception e) {}
        }
    }

  
    /**
     * Implements the search operation.  This performs a search on the index
     * in the current object based on the 'query' string parameter.
     */
    private void doReindex(DOOperationContext context, InputStream in, OutputStream out)
    throws Exception
    {
      HeaderSet headers = context.getOperationHeaders();
      String objectId = headers.getStringHeader("objectid",null);
      if(objectId==null) objectId = context.getTargetObjectID();
      try {
          // need to lock this since it happens outside the main indexing thread
          synchronized(indexWriterLock) {
              initializeIndexWriter();  
              updateObjectFromStorage(objectId);
              indexWriter.commit();
          }
          indexSearcherManager.reopen();
      } catch (Exception e) {
        sendErrorResponse(out, "There was an error reindexing object "+context.getTargetObjectID(),
                          DOException.STORAGE_ERROR);
        return;
      }
      sendSuccessResponse(out);
    }  

// Not a terrible idea, but doesn't work due to possibility of multiple transactions in a single millisecond.  
// Would have to include transaction keys in the index to get it robust.
//    private boolean isUpToDate(String objectId) throws IOException {
//        if(repoStorage==null) {
//            repoStorage = new ConcreteStorageProxy(serverMain.getStorage(), 
//                    serverMain.getServerID(),
//                    serverMain.getServerID(),
//                    null);
//        }
//        StorageProxy storage = repoStorage.getObjectAccessor(objectId);
//        
//        Query query = new TermQuery(new Term("id",objectId));
//        
//        IndexSearcherManager.LockTrackingIndexSearcherWrapper indexSearcherWrapper = indexSearcherManager.getAndLockIndexSearcherWrapper();
//        try {
//            IndexSearcher indexSearcher = indexSearcherWrapper.getIndexSearcher();
//
//            TopDocs hits = indexSearcher.search(query,1);
//            if(hits.totalHits==0) return !repoStorage.getObjectAccessor(objectId).doesObjectExist();
//            
//            Document doc = indexSearcher.doc(hits.scoreDocs[0].doc);
//            String indexDate = doc.get("objmodified");
//            String storageDate = DateTools.dateToString(new java.util.Date(storage.getAttributes(null).getLongHeader(DOConstants.DATE_MODIFIED_ATTRIBUTE, 0)),DateTools.Resolution.MILLISECOND);
//            return storageDate.equals(indexDate);
//        }
//        finally {
//            indexSearcherWrapper.releaseAndCloseIfNeeded();
//        }
//    }
    
    private void initializeIndexWriter() throws IOException {
        if(indexWriter==null) {
            // double-checked locking works with volatile
            synchronized(indexWriterLock) {
                if(indexWriter==null) {
                    indexWriter = new IndexWriter(indexDirectory, indexBuilder.getAnalyzer(), false, IndexWriter.MaxFieldLength.UNLIMITED);
                }
            }
        }
    }

  
    /**
     * Implements the search operation.  This performs a search on the index
     * in the current object based on the 'query' string parameter.
     */
    private void doSearch(DOOperationContext context, InputStream in, OutputStream out)
    throws Exception
    {
        StorageProxy storage = context.getStorage();
        HeaderSet params = context.getOperationHeaders();
        String queryStr = params.getStringHeader("query", null);
        if(queryStr==null) {
            sendErrorResponse(out, "Request was missing 'query' parameter",
                    DOException.APPLICATION_ERROR);
            return;
        }        
        
        String returnedFieldsStr = params.getStringHeader("returnedFields", null);
        List<String> returnedFields = null; 
        if (returnedFieldsStr!= null) returnedFields = getFieldsFromString(returnedFieldsStr);
        
        String sortFieldsStr = params.getStringHeader("sortFields", null); //sortFieldsStr is not yet used
        List<String> sortFields = null; 
        if (sortFieldsStr!= null) sortFields = getFieldsFromString(sortFieldsStr);
                
        String sortOrder = params.getStringHeader("sortOrder", "ascending"); 
        if (sortOrder == null) sortOrder = "ascending";
        int pageSize = params.getIntHeader("pageSize", 0);
        int pageOffset = params.getIntHeader("pageOffset", 0);

        boolean unsupportedFederationParams = (returnedFieldsStr!=null) || (sortFieldsStr!=null) || (pageSize!=0);
        boolean doFederation = federatedSearchTargets!=null && federatedSearchTargets.size()>0 && 
        		params.getBooleanHeader("federation", true);        
        
        if(doFederation && unsupportedFederationParams) {
        	sendErrorResponse(out, "Federated search does not support the following parameters: pageSize, pageOffset, sortOrder, sortFields, returnedFields", DOException.OPERATION_NOT_AVAILABLE);
        	return;
        }

        if(DEBUG) System.err.println("searching lucene for: "+queryStr);

        Query q0;
        try {
            q0 = indexBuilder.getQueryParser().parse(queryStr);
        }
        catch(ParseException e) {
            sendErrorResponse(out,"Parse failure: " + queryStr, DOException.APPLICATION_ERROR);
            return;
        }

        //q0.setBoost(10.0f);
        //BooleanQuery query = new BooleanQuery();
        //query.add(q0, BooleanClause.Occur.MUST);

        ArrayList federatedThreads = new ArrayList();
        IndexSearcherManager.LockTrackingIndexSearcherWrapper indexSearcherWrapper = indexSearcherManager.getAndLockIndexSearcherWrapper();
        try {
            IndexSearcher indexSearcher = indexSearcherWrapper.getIndexSearcher();

            if(DEBUG) System.err.println("search for <<<"+q0+">>> index-size:"+indexSearcher.getIndexReader().numDocs());

            TopDocs hits = indexSearcher.search(q0,MAX_DOCS);
            if(DEBUG) System.err.println("Found "+hits.totalHits+" results.");
            sendSuccessResponse(out);

            List<ScoreDoc> results = Arrays.asList(hits.scoreDocs);
            results = limitResultsToPermitted(results, context, indexSearcher);

            if(!doFederation) {
            	HeaderSet queryMetaData = new HeaderSet("queryInfo");
            	queryMetaData.addHeader("totalMatches",results.size());
            	queryMetaData.writeHeaders(out);            
            }
            	
            if(doFederation) {
                // perform federated search, returning results from searches on other repositories
                params.removeHeadersWithKey("federation");
                params.addHeader("federation", false);

                for(Iterator iter=federatedSearchTargets.iterator(); iter.hasNext(); ) {
                    FederatedSearcher fedSearch = new FederatedSearcher((String)iter.next(), context, params, out);
                    fedSearch.start();
                    federatedThreads.add(fedSearch);
                }
            }
            
            HeaderSet result = new HeaderSet("result");
            if (sortFields != null) { //Lucene cannot sort on tokenized fields so the sort is performed post search
            	results = sort(results, sortFields, indexSearcher);
            	if ("descending".equals(sortOrder)) {
            		Collections.reverse(results);
            	} 
            } 
            if (pageSize > 0) {
            	results = limitResultsToPage(results, pageSize, pageOffset);
            }
            
            for (ScoreDoc scoreDoc : results) {
                result.removeAllHeaders();
                Document doc = indexSearcher.doc(scoreDoc.doc);
                String objectID = doc.get("id");
                if(objectID==null) continue;
                result.addHeader("objectid", objectID);
                result.addHeader("repoid", serverMain.getServerID());
                result.addHeader("score", String.valueOf(scoreDoc.score));
                for(Fieldable f : doc.getFields()) {
                	if (returnedFields==null) {
                		addField(result, f);
                	}
                	else if (returnedFields.contains(f.name())) { 
                		addField(result, f);
                	}
                	else if(escapedElementAttContains(returnedFields,f.name())) {
                        addField(result, f);
                	}
                }
                synchronized(out) {
                    result.writeHeaders(out);
                }
            }
        }
        finally {
            indexSearcherWrapper.releaseAndCloseIfNeeded();
        }

        // wait for the federated searches to finish
        for(Iterator iter=federatedThreads.iterator(); iter.hasNext(); ) {
            ((Thread)iter.next()).join();
        }

        out.close();
    }

    // A bit of a hack to allow users to specify unescaped names and get escaped names returned
    private static boolean escapedElementAttContains(List<String> returnedFields, String name) {
        if(name==null) return false;
        if(!name.startsWith("el")) return false;
        int underscore1 = name.indexOf("_");
        if(underscore1 < 0) return false;
        int underscore2 = name.indexOf("_",underscore1+1);
        if(underscore2 < 0) return false;
        String unescapedName = name.substring(0,underscore1) + StringUtils.decodeURLIgnorePlus(name.substring(underscore1,underscore2)) + name.substring(underscore2);
        return returnedFields.contains(unescapedName);
    }
    
	private void addField(HeaderSet result, Fieldable f) {
		String val = f.stringValue();
		if(val==null) val = "";
		result.addHeader("field:"+f.name(), val);
	}
	
	//Filters out those search results that this context is not permitted to access
	private List<ScoreDoc> limitResultsToPermitted(List<ScoreDoc> results, DOOperationContext context, IndexSearcher indexSearcher) 
		throws IOException {
		List<ScoreDoc> permittedResults = new ArrayList<ScoreDoc>();
		net.cnri.apps.doserver.Authorizer authorizer = serverMain.getAuthorizer();
		for (ScoreDoc scoreDoc : results) {
            Document doc = indexSearcher.doc(scoreDoc.doc);
            String objectID = doc.get("id");
            if(objectID==null) continue;
            if(authorizer.operationIsAllowed(context, objectID, DOConstants.GET_DATA_OP_ID)) {
            	permittedResults.add(scoreDoc);
            }
		}
		return permittedResults;
	}
    
	//returns a limited list of the results. pageOffset begins at zero.
	private List<ScoreDoc> limitResultsToPage(List<ScoreDoc> results, int pageSize, int pageOffset) {
		int fromIndex = pageSize*pageOffset;
		int toIndex = fromIndex+pageSize;
		return limitResults(results, fromIndex, toIndex);
	}
	
	//fromIndex is inclusive, toIndex is exclusive.
	private List<ScoreDoc> limitResults(List<ScoreDoc> results, int fromIndex, int toIndex) {
		List<ScoreDoc> resultsPage = results.subList(fromIndex, toIndex);
		return resultsPage;
	}   
    
    private List<String> getFieldsFromString(String s) {
    	return Arrays.asList(s.split(","));
    }
    
    private class FederatedSearcher
    extends Thread
    {
        private String searchTarget;
        private DOOperationContext context;
        private HeaderSet params;
        private OutputStream out;

        FederatedSearcher(String searchTarget, DOOperationContext context, HeaderSet params,
                OutputStream out) {
            this.searchTarget = searchTarget;
            this.context = context;
            this.params = params;
            this.out = out;
        }

        public void run() {
            try {
                if(DEBUG) System.err.println("federating search to "+searchTarget);
                context.performOperation(null, searchTarget, SEARCH_OP_ID, params,
                        new ByteArrayInputStream(new byte[0]), 
                        new DontCloseOutputStream(this, out));
                if(DEBUG) System.err.println("finished search federation to "+searchTarget);
            } catch (Exception e) {
                System.err.println("Error federating search to index: "+searchTarget+": "+e);
                e.printStackTrace(System.err);
            }
        }

        public String toString() {
            return "fedsearch at '"+searchTarget+"' for terms '"+params+"'";
        }

        void receivedResult(byte line[]) 
        throws IOException
        {
            synchronized(out) { out.write(line); }
            if(DEBUG) System.err.println("result from "+searchTarget+":  "+new String(line));
        }
    }
    
    //Sorting the results post search rather than using the Lucene sort as Lucene cannot sort on tokenized fields.
    private List<ScoreDoc> sort(List<ScoreDoc> scoreDocs, List<String> sortFields, IndexSearcher indexSearcher) {
    	List<ScoreDoc> result = new ArrayList<ScoreDoc>();
    	List<SortableScoreDoc> sortableScoreDocs = new ArrayList<SortableScoreDoc>();
    	for (ScoreDoc scoreDoc : scoreDocs) {
    		Document doc = null;
    		try {
    			doc = indexSearcher.doc(scoreDoc.doc);
    			sortableScoreDocs.add(new SortableScoreDoc(scoreDoc, doc, sortFields));
    		} catch (IOException ioe){
    			ioe.printStackTrace(System.out);
    		}
    	}
    	Collections.sort(sortableScoreDocs);
    	for (SortableScoreDoc ssd : sortableScoreDocs) {
    		result.add(ssd.scoreDoc);
    	}
    	return result;
    }

    private class SortableScoreDoc implements Comparable<SortableScoreDoc>{
    	private ScoreDoc scoreDoc;
    	private Document doc;
    	private HashMap<String, String> fields;
    	private List<String> sortFields;
    	    	
    	public SortableScoreDoc(ScoreDoc scoreDoc, Document doc, List<String> sortFields) {
    		this.scoreDoc = scoreDoc;
    		this.doc = doc;
    		this.fields = getFields();
    		this.sortFields = sortFields;
    	}
        
    	public int compareTo(SortableScoreDoc other) {
    		for (String sortField : sortFields) {
    			String thisField = get(sortField);
    			String otherField = other.get(sortField);
    			if ((thisField != null)&&(otherField == null)) return 1;  
    			if ((thisField == null)&&(otherField != null)) return -1; 
    			if ((thisField == null)&&(otherField == null)) continue; //compare the next field
    			if (thisField.compareTo(otherField) == 0) continue; //compare the next field
    			return thisField.compareTo(otherField);
    		}
    		return 0; 
    	}
    	
    	public String get(String key) {
    		return fields.get(key);
    	}
    	
    	private HashMap<String, String> getFields() {
    		HashMap<String, String> result = new HashMap<String, String>();
    		result.put("do_id", doc.get("id"));
    		result.put("score", String.valueOf(scoreDoc.score));
    		for(Fieldable f : doc.getFields()) {
    			result.put(f.name(), f.stringValue());
    		}
    		return result;
    	}
    }

    private class DontCloseOutputStream
    extends OutputStream
    {
        private OutputStream out;
        private ByteArrayOutputStream bout;
        private boolean sawStatusLine = false;
        private FederatedSearcher searcher;

        DontCloseOutputStream(FederatedSearcher searcher, OutputStream out) {
            this.searcher = searcher;
            this.out = out;
            this.bout = new ByteArrayOutputStream();
        }

        public void write(int b) 
        throws IOException
        {
            bout.write(b);
            if(((byte)b)=='\n') {
                byte line[] = bout.toByteArray();
                // TODO: Check for first line, which should be the operation's status and skip it
                if(!sawStatusLine) {
                    if(DEBUG) System.err.println("got status from search: "+net.handle.hdllib.Util.decodeString(line));
                    sawStatusLine = true;
                } else {
                    searcher.receivedResult(line);
                }
                bout.reset();
            }
        }

        public void close() {
            // noop
        }
    }


    /** Updates the indexes from any sources specified in the config file */
    private void updateIndex(boolean bigUpdate) {
        boolean txnUpdateStillNeeded = txnUpdateNeeded;
        StreamTable indexConfig = null;
        try {
            // check for the existence of the "suppress_indexing" file
            if(suppressIndexingFile.exists()) return;

            if(bigUpdate) {
                // (re-)initialize the authentication used for indexing
                CollaborationHub hub = serverMain.getServiceRegistry();
                Object keyrings[] = hub.lookupEntity(null, "lucene-indexing-keyring", DOKeyRing.class);
                if(keyrings!=null && keyrings.length>0) {
                    indexBuilder.setKeyRing((DOKeyRing)keyrings[0]);
                } else {
                    indexBuilder.setKeyRing(null);
                }
            }

            try {
                initializeIndexWriter();
            } catch (IOException e) {
                System.err.println("Error opening index writer: "+e);
                e.printStackTrace(System.err);
                return;
            }

            if(bigUpdate) {
                InputStream in = null;
                indexConfig = new StreamTable();
                try {
                    if(repoStorage==null) {
                        repoStorage = new ConcreteStorageProxy(serverMain.getStorage(), 
                                serverMain.getServerID(),
                                serverMain.getServerID(),
                                null);
                    }
                    if(repoStorage.doesDataElementExist(LUCENE_CONFIG_ELEMENT)) {
                        in = repoStorage.getDataElement(LUCENE_CONFIG_ELEMENT);
                    }
                    if(in!=null) {
                        indexConfig.readFrom(in);

                        Object fedVect = indexConfig.get(FEDERATED_INDEXES_KEY);
                        if(fedVect!=null && fedVect instanceof List) {
                            federatedSearchTargets = (List)fedVect;
                        } else {
                            federatedSearchTargets = new ArrayList();
                        }
                    } else {
                        federatedSearchTargets = new ArrayList();
                    }
                } catch (Exception e) {
                    System.err.println("Error reading configuration: "+e);
                    e.printStackTrace(System.err);
                    return;
                } finally {
                    try { in.close(); } catch (Exception e) {}
                }
            }    
            txnUpdateNeeded = false;
            txnUpdateStillNeeded = false;
        }
        finally {
            if(txnUpdateStillNeeded) {
                try {
                    Thread.sleep(30 * 1000);
                }
                catch(InterruptedException e) {
                    Thread.currentThread().interrupt();
                }
            }
        }

        Exception indexWriterException = null;
        try {
            // retrieve any updates from the repository in which we live
            updateFromContainer();
            try {
                indexWriter.commit();
            }
            catch(IOException e) {
                indexWriterException = e;
                System.err.println("Error committing index: "+e);
            }

            if(bigUpdate) {
                // retrieve any updates from other repositories
                StreamTable config = (StreamTable)indexConfig.deepClone();
                StreamVector sources = (StreamVector)config.get("sources");
                for(int i=0; sources!=null && i<sources.size(); i++) {
                    updateFromSource(String.valueOf(sources.get(i)));
                }

                // scan objects that other indexes couldn't scan in the hopes that we 
                // have the necessary rights to access and decrypt those objects
                StreamVector sourceIndexes = (StreamVector)indexConfig.get("srcindexes");
                for(int i=0; sourceIndexes!=null && i<sourceIndexes.size(); i++) {
                    updateFromIndex(String.valueOf(sourceIndexes.get(i)));
                }

                try {
                    indexWriter.optimize();
                    indexWriter.commit();
                } catch (IOException e) {
                    indexWriterException = e;
                    System.err.println("Error optimizing and committing index: "+e);
                }
            }
        } finally {
            if(indexWriterException!=null) {
                try {
                    synchronized(indexWriterLock) {
                        indexWriter.close();
                        indexWriter = null;
                    }
                } catch (Exception e) {
                    System.err.println("Error closing index after update: "+e);
                    e.printStackTrace(System.err);
                }
            }
        }
        indexSearcherManager.reopen();
    } 
    

    /** Updates the index from the given source */
    private void updateFromIndex(String indexID) {
        if(DEBUG) System.err.println("Updating from index: "+indexID);
        StreamTable status = (StreamTable)indexScanStatus.get("index:"+indexID);
        synchronized(indexScanStatus) {
            if(status==null) {
                status = new StreamTable();
                indexScanStatus.put("index:"+indexID, status);
            }
        }
        DOClient client = new DOClient(auth);
        HashMap repos = new HashMap();

        HeaderSet params = new HeaderSet();
        String lastTimeStamp = status.getStr("last_idx_timestamp", null);
        StringBuffer query = new StringBuffer();
        query.append("idxerror:").append(DOException.PERMISSION_DENIED_ERROR).append(' ');
        if(lastTimeStamp!=null) {
            query.append("dtindexed:[").append(lastTimeStamp).append(" TO 999999999999999999999] ");
        }
        params.addHeader("query", query.toString());

        DOAuthentication localAuth = auth;
        if(localAuth==null) {
            localAuth = PKAuthentication.getAnonymousAuth();
        }

        StreamPair io = null;
        try {
            io = client.performOperation(indexID, SEARCH_OP_ID, params);
            HeaderSet response = new HeaderSet();
            InputStream in = io.getInputStream();

            while(response.readHeaders(in)) {
                String objectID = response.getStringHeader("objectid", null);
                if(objectID==null) continue;
                String repoID = response.getStringHeader("repoid", null);
                System.err.println("indexing restricted record: "+objectID+" @ "+repoID);

                if(repoID==null) {
                    repoID = DOClient.resolveRepositoryID(objectID);
                }

                Repository repo = (Repository)repos.get(repoID);
                if(repo==null) {
                    repo = new Repository(localAuth, repoID);
                    repos.put(repoID, repo);
                }

                updateObjectInIndex(repo, objectID);

                String lastIndexed = response.getStringHeader("field:dtindexed", null);
                if(lastIndexed!=null) {
                    status.put("last_idx_timestamp", lastIndexed.trim());
                }
            }
        } catch (Exception e) {
            System.err.println("Error updating from index "+indexID+": "+e);
            e.printStackTrace(System.err);
        } finally {
            try { io.close(); } catch (Throwable t) {}
            saveSourceStatus();  // save the source/index status info
        }

    }




    private void updateFromContainer() {
        HashMap updatedObjects = new HashMap();

        try {
            long lastTimestamp = indexScanStatus.getLong("self_scan_txn", -1);
            ArrayList objectsToUpdate = new ArrayList();

            if(lastTimestamp==-1) {
                // we haven't done an update yet, scan all the objects in the repository
                // scan all of the objects...
                Enumeration allObjects = repoStorage.listObjects();
                while(allObjects.hasMoreElements()) {
                    updateObjectFromStorage((String)allObjects.nextElement());
                }

                lastTimestamp = 0;
                // get the most recent timestamp, unfortunately by scanning the entire queue (at least it is local)
                Enumeration txnEnum = serverMain.getTxnQueue().getScanner(lastTimestamp);
                while(txnEnum.hasMoreElements()) {
                    Transaction txn = (Transaction)txnEnum.nextElement();
                    lastTimestamp = txn.timestamp;
                }
            } else {
                // we are updating all objects since the last transaction update
                Enumeration txnEnum = serverMain.getTxnQueue().getScanner(lastTimestamp);
                while(txnEnum.hasMoreElements()) {
                    Transaction txn = (Transaction)txnEnum.nextElement();

                    if(txn.objectID==null) continue; // ignore all non-object specific transactions

                    switch(txn.action) {
                    case Transaction.ACTION_COMMENT:
                        break;
                    case Transaction.ACTION_OBJ_DEL:
                    case Transaction.ACTION_OBJ_ADD:
                    case Transaction.ACTION_OBJ_DATA_UPDATE:
                    case Transaction.ACTION_OBJ_DATA_DELETE:
                    case Transaction.ACTION_OBJ_ATTRIBUTE_UPDATE:
                    case Transaction.ACTION_OBJ_ATTRIBUTE_DELETE:
                        if(!updatedObjects.containsKey(txn.objectID)) {
                            // only update the object once per session
                            objectsToUpdate.add(txn.objectID);
                            updatedObjects.put(txn.objectID, "");
                        }
                        break;
                    }

                    lastTimestamp = txn.timestamp;

                    if(objectsToUpdate.size() >= NUM_OBJECTS_BATCH_SIZE) {
                        break;
                    }
                }

                for(Iterator iter=objectsToUpdate.iterator(); iter.hasNext(); ) {
                    String objectID = (String)iter.next();
                    if (DEBUG) System.err.println("updating object: " + objectID + " from storage");
                    updateObjectFromStorage(objectID);
                }
            }

            indexScanStatus.put("self_scan_txn", lastTimestamp);
        } catch (Exception e) {
            System.err.println("Error trying to get txns from storage: "+e);
            if(DEBUG) e.printStackTrace(System.err);
        } finally {
            saveSourceStatus();  // save the source/index status info
        }
    }



    /** Updates the index from the given source */
    private void updateFromSource(String sourceID) {
        StreamTable sourceStatus;

        DOClientConnection connection = null;
        DOServiceInfo svcInfo = null;
        if(DEBUG) System.err.println("updating from "+sourceID);
        try {
            svcInfo = new DOServiceInfo(sourceID);
        } catch (Exception e) {
            System.err.println("Error resolving source: "+sourceID+"; error: "+e);
            return;
        }

        synchronized(indexScanStatus) {
            sourceStatus = (StreamTable)indexScanStatus.get(sourceID);
            if(sourceStatus==null) {
                indexScanStatus.put(sourceID, sourceStatus = new StreamTable());
            }
        }


        for(int i=0; i<svcInfo.getServerCount(); i++) {
            DOServerInfo svrInfo = svcInfo.getServer(i);
            if(svrInfo==null) continue;

            String serverID = svrInfo.getServerID();
            StreamTable serverStatus = null;
            synchronized(sourceStatus) {
                serverStatus = (StreamTable)sourceStatus.get(serverID);
                if(serverStatus==null) {
                    sourceStatus.put(serverID, serverStatus = new StreamTable());
                }
            }

            if(DEBUG) System.err.println("updating from: src="+sourceID+"; svr="+serverID);
            HeaderSet params = new HeaderSet();
            if(serverStatus.containsKey("last_txn_id")) {
                params.addHeader("txn_id", serverStatus.getStr("last_txn_id", null));
            }

            StreamPair io = null;
            Repository repo = null;
            DOClientConnection conn = null;
            try {
                DOAuthentication localAuth = auth;
                if(localAuth==null) {
                    localAuth = PKAuthentication.getAnonymousAuth();
                }
                conn = new DOClientConnection(localAuth);
                conn.setUseEncryption(false);
                conn.connect(svrInfo);

                repo = new Repository(sourceID, conn);

                io = conn.performOperation(sourceID, DOConstants.GET_REPO_TXNS_OP_ID,
                        params);

                InputStream in = io.getInputStream();
                HeaderSet record = new HeaderSet();
                HashMap updatedObjects = new HashMap();

                while(record.readHeaders(in)) {
                    String timestampStr = record.getStringHeader("tstamp", null);
                    if(timestampStr==null) {
                        System.err.println("Warning:  ignoring received transaction with "+
                                "null timestamp: "+record);
                        continue;
                    }

                    String objID = record.getStringHeader("object_id", null);
                    if(objID==null) continue; // ignore all non-object specific transactions
                    long timestamp = record.getLongHeader("tstamp", 0);

                    String txnType = record.getStringHeader(PARAM_TXN_TYPE, "");
                    if(txnType.equals(TXN_TYPE_COMMENT)) {
                        // nothing to index...
                    } else if(txnType.equals(TXN_TYPE_DEL_OBJ)) {
                        // delete the object from the index
                        deleteObjectFromIndex(objID);
                    } else {
                        // for all other modifications update the object in the index

                        if(!updatedObjects.containsKey(objID)) {
                            // only update the object once per session
                            updateObjectInIndex(repo, objID);
                            updatedObjects.put(objID, "");
                        }
                    }

                    serverStatus.put("last_txn_id", timestamp);
                }
            } catch (Exception e) {
                System.err.println("Error trying to get txns from src="+sourceID+
                        " svr="+serverID+": "+e);
                if(DEBUG) e.printStackTrace(System.err);
            } finally {
                try { io.close(); } catch (Throwable t) {}
                saveSourceStatus();  // save the source status info
                try {
                    conn.close();
                } catch (Throwable t) {
                    System.err.println("Warning: Error closing connection to "+sourceID+": "+t);
                }
                conn = null;
            }
        }
    }


    private void saveSourceStatus() {
        try {
            synchronized(indexScanStatus) {
                indexScanStatus.writeToFile(scanStatusFile);
            }
        } catch (Exception e) {
            System.err.println("Error saving replication status: "+e);
            if(DEBUG) e.printStackTrace(System.err);
        }
    }

    /** Re-retrieve the object and add it to the index */
    private void deleteObjectFromIndex(String objID) 
    throws IOException
    {
        if(DEBUG) System.err.println("deleting object: "+objID);

        indexWriter.deleteDocuments(new Term("id", objID));
    }


    /** Update the object directly from the storage module */
    private void updateObjectFromStorage(String objectID) 
    throws Exception
    {
        if(DEBUG) System.err.println("Indexing object from storage: "+objectID);
        Thread thread = Thread.currentThread();
        String oldThreadName = thread.getName();
        thread.setName("Indexing object '"+objectID+"' directly from storage");
        try {
            StorageProxy objStore = repoStorage.getObjectAccessor(objectID);
            if(!objStore.doesObjectExist()) {
                indexWriter.deleteDocuments(new Term("id", objectID)); // delete any old entries for this ID
                return;
            }
            Document doc = indexBuilder.documentOfStorageProxy(objStore);
            indexWriter.deleteDocuments(new Term("id", objectID)); // delete any old entries for this ID
            indexWriter.addDocument(doc);  // add the new document
        } finally {
            thread.setName(oldThreadName);
        }
    }

    /** Re-retrieve the object and add it to the index */
    private void updateObjectInIndex(Repository repo, String objectID) 
    throws IOException
    {
        DigitalObject obj = repo.getDigitalObject(objectID);
        Document doc = indexBuilder.documentOfDigitalObject(obj);
        indexWriter.deleteDocuments(new Term("id", objectID)); // delete any old entries for this ID
        indexWriter.addDocument(doc);  // add the new document
    }

    private void sendErrorResponse(OutputStream out, String msg, int code)
    throws IOException
    {
        HeaderSet response = new HeaderSet("response");
        response.addHeader("status", "error");
        if(msg!=null)
            response.addHeader("message", msg);
        if(code>=0)
            response.addHeader("code", code);
        response.writeHeaders(out);
        out.flush();
    }

    private void sendSuccessResponse(OutputStream out)
    throws IOException
    {
        HeaderSet response = new HeaderSet("response");
        response.addHeader("status", "success");
        response.writeHeaders(out);
        out.flush();
    }


}


